# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Event implentation"""
from enum import Enum
from datetime import datetime

from .utils import get_id_gen


class Event(object):
    """Event class

    :param name: name of the event
    :param dt: occur datetime
    :param expire: expire in second
    :param event_id: id of the event
    :param \*\*kwargs: special kwargs for the event, can get by kwargs attribute

    can compare less/greater than with other event object by time, equal by event id.
    """
    id_gen = get_id_gen(prefix='Event')

    def __init__(self, env, name, dt=None, expire=0, event_id=None, **kwargs):
        self.env = env
        self._name = name
        self._event_id = event_id
        self._kwargs = kwargs
        self._expire = expire

        if dt:
            self._timestamp = dt.timestamp()
        else:
            self._timestamp = self.env.dispatcher.now.timestamp()
        if self._event_id is None:
            self._event_id = next(Event.id_gen)

    @property
    def id(self):
        """ID of the event"""
        return self._event_id

    @property
    def name(self):
        """Name of the event"""
        return self._name

    @property
    def timestamp(self):
        """Occur time of the event, in POSIX timestamp format"""
        return self._timestamp

    @property
    def expire(self):
        """expire expire in seconds after event occur time (timestamp attribute), 0 is no expire."""
        return self._expire

    @property
    def kwargs(self):
        """Parameters of the event, a `dict`"""
        return self._kwargs

    def __getattr__(self, name):
        return self._kwargs.get(name, None)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __lt__(self, other):
        if self.timestamp < other.timestamp:
            return True
        return False

    def __gt__(self, other):
        if self.timestamp > other.timestamp:
            return True
        return False

    def __repr__(self):
        return "<Event.%s:id=%s, time=%s, kwargs=%s>" % (self.name,
                                                         self.id,
                                                         datetime.fromtimestamp(self.timestamp),
                                                         self.kwargs)


class SysEvents(object):
    """system defined events

    enum of events that finchan system generate/reserved:

        * SYSTEM_STARTED = 'system_event.system_started'
        * SYSTEM_EXITING = 'system_event.system_will_exit'
    """
    SYSTEM_STARTED = 'system_event.system_started'
    SYSTEM_EXITING = 'system_event.system_will_exit'



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""time event source"""
import time
import logging

from finchan.event import SysEvents

# name of the extension
ext_name = 'finchan.exts.timer_callback'
# required extension
required_exts = ['finchan.exts.timer_source']

logger = logging.getLogger(__name__)

logger = logging.getLogger()


def timer_call(env):
    logger.info(f'timer callback @ {env.now}')


def add_timer(event):
    scheduler = event.env.get_ext_obj('finchan.exts.timer_source')
    if event.env.run_mode == 'live_track':
        scheduler.run_every(3).to(5).seconds().do(timer_call)
    else:
        scheduler.run_every(3).to(15).minutes().do(timer_call)
    logger.info('#TimerCall timer job added')


def load_finchan_ext(env, *args, **kwargs):
    env.dispatcher.subscribe(SysEvents.SYSTEM_STARTED, add_timer)

def unload_finchan_ext(env):
    scheduler = env.get_ext_obj('finchan.exts.timer_source')
    scheduler.cancel()


if __name__ == '__main__':
    # tests
    pass



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__title__ = 'finchan'
__description__ = 'An event process framework with python'
__url__ = 'https://github.com/qytz/finchan'
__version__ = '0.2.0'
__build__ = 0x000200
__author__ = 'qytz'
__author_email__ = 'hhhhhf@foxmail.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2017-present qytz'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_finchan
----------------------------------

Tests for `finchan` module.
"""


import sys
import unittest
from contextlib import contextmanager
from click.testing import CliRunner

import finchan
from finchan.__main__ import main



class TestFinchan(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_000_something(self):
        pass

    def test_command_line_interface(self):
        runner = CliRunner()
        result = runner.invoke(main)
        help_result = runner.invoke(main, ['--help'])
        assert help_result.exit_code == 0
        assert '--help             Show this message and exit.' in help_result.output



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-Present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EventSource Interface"""
import abc


class AbsEventSource(abc.ABC):
    """EventSource interface

    EventSource simulate some things in the real world to events,
    when event occurs it put the event into the event queue.

    main interface for dispatcher:

    :param start: start the event source.
    :param stop: stop the event source.
    :param gen_events: generate events and put them into event queue.
    """
    _name = None

    def __init__(self, *args, **kwargs):
        """init the event source"""
        pass

    @property
    def name(self):
        """Name of the event_source"""
        return self._name

    @abc.abstractmethod
    async def gen_events(self, event_queue, limit_dt=None):
        """Generate events for `Dispatcher`

        put all the generated/received events to event queue by call `Dispatcher.eq_put` ,

        event_queue: put all the generated event to the event_queue
        limit_dt: for backtest, gen events before the limit_dt

        for live track, the event source run in parallel mode, and should not return.

        for backtrack, the event source run in serial mode,
        and must return after put all events before limit_dt to event queue.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def start(self):
        """initialize the event_source, start generate/receive events"""
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        """stop the event_source, stop generate/receive events"""
        raise NotImplementedError



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import json
import asyncio
import logging
import datetime
import logging.config

import click
import uvloop

from .env import Env
from .exts import ExtManager
from .options import load_configs
from .dispatcher import BackTrackDispatcher, LiveDispatcher


@click.command()
@click.option(
    "-v", "--verbose", count=True, help="Count output level, can set multipule times."
)
@click.option("-c", "--config", help="Specify config file.")
def main(verbose=0, config=None):
    """Console script for finchan

    Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Project url: https://github.com/qytz/finchan
    """
    env = Env()
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    env.verbose = verbose
    if not config:
        conf_path = os.path.expanduser("~/.finchan/config.yml")
    else:
        conf_path = config
    try:
        env.options = load_configs(conf_path)
    except (SyntaxError, TypeError) as e:
        print("Parse configure file failed, please check: %s" % e)
        return

    work_dir = os.path.expanduser(env.options.get("work_dir", "~/.finchan"))
    os.makedirs(work_dir, exist_ok=True)
    os.makedirs(os.path.join(work_dir, "logs"), exist_ok=True)
    os.chdir(work_dir)
    log_config = env.options.get("log_config", {})
    # patch the log filter parameters
    if "filters" in log_config and "finchan" in log_config["filters"]:
        log_config["filters"]["finchan"]["env"] = env
    logging.config.dictConfig(log_config)
    if env.options["run_mode"] == "backtrack":
        env.run_mode = "backtrack"
    else:
        env.run_mode = "live_track"

    root_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(asctime)s %(tracktime)s %(levelname)-8s %(message)s")
    )
    if verbose > 0:
        handler.setLevel("DEBUG")
    else:
        handler.setLevel("INFO")
    root_logger.addHandler(handler)

    root_logger.info("Run in %s mode", env.run_mode)
    if env.run_mode == "backtrack":
        dispatcher_config = env.options.get("dispatcher.backtrack", {})
        dispatcher = BackTrackDispatcher(env, **dispatcher_config)
        enabled_exts = env.options.get("enabled_backtrack_exts", [])
    else:
        dispatcher_config = env.options.get("dispatcher.live_track", {})
        dispatcher = LiveDispatcher(env, **dispatcher_config)
        enabled_exts = env.options.get("enabled_live_exts", [])

    extm_args = env.options["ext_manager"]
    if not extm_args:
        extm_args = {}
    ext_manager = ExtManager(env, **extm_args)
    env.set_dispatcher(dispatcher)
    env.set_ext_manager(ext_manager)

    env.load_exts(enabled_exts)
    env.run()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""options manager"""
from pathlib import Path
from ruamel.yaml import YAML

yaml = YAML(typ="safe", pure=True)
# yaml.default_flow_style = False


def parse_yaml_conf(conf_path):
    """parse yaml option file with finchan options.

    return a dict with all options in it.
    """
    try:
        return yaml.load(Path(conf_path))
    except Exception as e:
        raise SyntaxError("load configure file failed: %s" % str(e))


def merge_configs(a, b):
    """deep merge dict configs from b to a

    merge all dict&list values, recursely, will modify a object and return"""
    if type(a) != type(b):
        raise TypeError("Different type of merge objects: %s<=>%s" % (type(a), type(b)))

    if not isinstance(a, (list, dict)):
        raise TypeError("can only merge dict/list, get %s" % type(a))

    if isinstance(a, list):
        for val in b:
            if val not in a:
                a.append(val)
    else:
        for key, val in b.items():
            if key not in a:
                a[key] = val
                continue

            if not isinstance(a[key], type(val)):
                raise TypeError(
                    "Different types for key %s found: %s<=>%s"
                    % (key, type(a), type(b))
                )

            if isinstance(val, (list, dict)):
                merge_configs(a[key], val)
            else:
                raise TypeError(
                    "can only merge dict/list, get %s for key %s" % (type(val), key)
                )
    return a


def load_configs(conf_path):
    """load yaml configure files from a path, support one file and multiple configure files in the path"""
    p = Path(conf_path).expanduser()
    if p.is_file():
        return parse_yaml_conf(p)

    configs = [parse_yaml_conf(f) for f in p.rglob("*.yml")]

    merged_config = {}
    for config in configs:
        if not config:  # skip empty configure file
            continue
        merge_configs(merged_config, config)
    return merged_config



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-Present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""kvstore Interface, inspired by redis-py."""
import abc


class AbsKvStore(abc.ABC):
    """KV store interface"""
    _name = None

    def __init__(self, *args, **kwargs):
        """init the kvstore connection"""
        pass

    # --- kv ---
    def set(self, name, value):
        """Set the value at key name to value."""
        raise NotImplementedError

    def get(self, name):
        """Return the value at key name, or None if the key doesn’t exist."""
        raise NotImplementedError

    def mset(self, *args, **kwargs):
        """Sets key/values based on a mapping.
        Mapping can be supplied as a single dictionary argument or as kwargs."""
        raise NotImplementedError

    def mget(self, keys, *args):
        """Returns a list of values ordered identically to keys."""
        raise NotImplementedError

    def setnx(self, name, value):
        """Set the value of key name to value if key doesn’t exist."""
        raise NotImplementedError

    def delete(self, name):
        """Delete the key and its value."""
        raise NotImplementedError

    # --- hashmap ---
    def hset(self, name, key, value):
        """Set key to value within hash name.

        Returns 1 if HSET created a new field, otherwise 0."""
        raise NotImplementedError

    def hget(self, name, key):
        """Return the value of ``key`` within the hash ``name``."""
        raise NotImplementedError

    def hmset(self, name, mapping):
        """Set key to value within hash ``name`` for each corresponding
        key and value from the ``mapping`` dict."""
        raise NotImplementedError

    def hmget(self, name, keys, *args):
        """Returns a list of values ordered identically to ``keys``."""
        raise NotImplementedError

    def hgetall(self, name):
        """Return a Python dict of the hash's name/value pairs."""
        raise NotImplementedError

    def hsetnx(self, name, key, value):
        """Set key to value within hash name if key does not exist.

        Returns 1 if HSETNX created a field, otherwise 0."""
        raise NotImplementedError

    def hdel(self, name, *keys):
        """Delete ``keys`` from hash ``name``."""
        raise NotImplementedError

    def hexists(self, name, key):
        """Returns a boolean indicating if ``key`` exists within hash ``name``."""
        raise NotImplementedError

    def hkeys(self, name):
        """Return the list of keys within hash ``name``."""
        raise NotImplementedError

    def hvals(self, name):
        """Return the list of values within hash ``name``."""
        raise NotImplementedError

    def hlen(self, name):
        """Return the number of elements in hash ``name``."""
        raise NotImplementedError

    # --- list ---
    def lpop(self, name):
        """Remove and return the first item of the list ``name``."""
        raise NotImplementedError

    def lpush(self, name, value):
        """Push ``values`` onto the head of the list ``name``."""
        raise NotImplementedError

    def rpop(self, name):
        """Remove and return the last item of the list ``name``."""
        raise NotImplementedError

    def rpush(self, name, *value):
        """Push ``values`` onto the tail of the list ``name``."""
        raise NotImplementedError

    def lset(self, name, index, value):
        """Set ``position`` of list ``name`` to ``value``."""
        raise NotImplementedError

    def lrange(self, name, start, end):
        """Return a slice of the list ``name`` between
        position ``start`` and ``end``.

        ``start`` and ``end`` can be negative numbers just like
        Python slicing notation."""
        raise NotImplementedError

    def ltrim(self, name, start, end):
        """Trim the list ``name``, removing all values not within the slice
        between ``start`` and ``end``.

        ``start`` and ``end`` can be negative numbers just like
        Python slicing notation."""
        raise NotImplementedError

    def llen(self, name):
        """Return the length of the list ``name``"""
        raise NotImplementedError



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Event Dispatcher

get events from event source and dispatch evnets to the subscriber.
"""
import os
import abc
import time
import signal
import asyncio
import logging
import threading
from queue import Queue, PriorityQueue, Empty
from datetime import datetime
from collections import defaultdict

from .event import Event, SysEvents
from finchan.interface.event_source import AbsEventSource


logger = logging.getLogger(__name__)


class BaseDispatcher(abc.ABC):
    """Event dispatcher abstract clsss.

    methods subclass should implement:

        * setup: need to start all the event sources and can also do other setup things.
        * cleanup: stop all the evnent source and do other cleanup things.
        * get_next_event: return the next event to dispath.


    :param eventq: queue to r/s event
    :param end_dt: dispatch end time
    """
    def __init__(self, env, **kwargs):
        self.env = env
        self._stop_flag = False
        self._event_sources = []
        self._end_dt = kwargs.get('end_dt')
        self._eventq = kwargs.get('eventq')
        self._event_subscribes = defaultdict(list)

        self._start_dt = datetime.now()
        self._event_timestamp = self._event_process_timestamp = self._start_dt.timestamp()

        # signal: ctrl_c/ctrl_\ quit
        signal.signal(signal.SIGINT, self.quit_handler)
        # signal.signal(signal.SIGQUIT, self.quit_handler)

    @property
    def now(self):
        """datetime of current"""
        return datetime.now()

    def quit_handler(self, signum, frame):
        """quit signal handler, callbacks for signal: ctrl\_c/ctrl\_\\"""
        if not self._stop_flag:
            self._stop_flag = True
            logger.info('#Dispatcher get signal: %s system will exit.', signum)
        else:
            logger.info('#Dispatcher get signal: %s system is exiting, please wait...', signum)

    def register_event_source(self, event_source):
        """register new event source

        :param event_source: an `AbsEventSource` subclass instance object
        :return: True if register succeed else False.
        """
        if not isinstance(event_source, AbsEventSource):
            logger.warning('#Dispatcher Trying to add invalid EventSource falied: '
                           'EventSource must be isinstance of AbsEventSource')
            return False
        self._event_sources.append(event_source)
        logger.info('#Dispatcher New event source has been registered: %s.', event_source)
        return True

    def deregister_event_source(self, event_source):
        """deregister an event source

        :param event_source: an `AbsEventSource` subclass instance object
        :return: True if deregister succeed, else False.
        """
        if not isinstance(event_source, AbsEventSource):
            logger.warning('#Dispatcher Trying to remove invalid EventSource falied: '
                           'EventSource must be isinstance of AbsEventSource')
            return False
        self._event_sources.remove(event_source)
        logger.info('#Dispatcher Event source %s has been deregistered.', event_source)
        return True

    def subscribe(self, name, callback, prepend=False):
        """subscribe an event

        when the event occurs, the callback will be called.

        :param name: name of the event
        :param callback: the callback func will be called
        :param prepend: add to the head of callbacks if `True` , else add to the end.
        :return: True if subscribe succeed, else False.
        """
        if not callable(callback):
            logger.warning(
                '#Dispatcher Failed to subscribe event: callback callable must be callable.')
            return False
        if not isinstance(name, str):
            logger.warning('#Dispatcher Failed to subscribe event: event name must be str.')
            return False
        self._event_subscribes.setdefault(name, [])
        if prepend:
            self._event_subscribes[name].insert(0, callback)
        else:
            self._event_subscribes[name].append(callback)
        logger.info('#Dispatcher New callback: %s for event %s has been added.', callback, name)
        return True

    def unsubscribe(self, name, callback):
        """unsubscribe an event

        :param name: name of the event
        :param callback: the callback func will be called
        :return: True if unsubscribe succeed, else False.
        """
        if not callable(callback):
            logger.warning(
                '#Dispatcher Failed to subscribe event: callback callable must be callable.')
            return False
        if not isinstance(name, str):
            logger.warning('#Dispatcher Failed to subscribe event: event name must be str.')
            return False
        self._event_subscribes.setdefault(name, [])
        self._event_subscribes[name].remove(callback)
        logger.info('#Dispatcher Callback: %s for event %s has been removed.', callback, name)
        return True

    # --- event_queue op ---
    def eq_put(self, *events):
        """put an event into event_queue, should not be blocked"""
        for event in events:
            if not isinstance(event, Event):
                logger.warning(
                    '#Dispatcher Put event failed: event must be instance of Event, %s', event)
                continue
            self._eventq.put(event)

    def eq_get(self, timeout=None):
        """Remove and return an event from the queue

        May be blcoked if the event_queue is empty.
        """
        event = self._eventq.get(timeout=timeout)
        if event:
            self._eventq.task_done()
        return event

    def eq_get_nowait(self):
        """Remove and return an event from the queue

        return None if the event_queue is empty.
        """
        event = self._eventq.get_nowait()
        if event:
            self._eventq.task_done()
        return event

    def eq_size(self):
        """count of event in the queue"""
        return self._eventq.qsize()

    def es_size(self):
        """count of `EventSource` that registered"""
        return len(self._event_sources)

    def cb_size(self, event):
        """count of callbcaks of the event"""
        return len(self._event_subscribes[event])

    def start_event_sources(self):
        """start the event sources"""
        for event_source in self._event_sources:
            try:
                logger.info('#Dispatcher starting event source %s...', event_source)
                event_source.start()
            except Exception as e:
                logger.warning('#Dispatcher Start event source %s failed: %s', event_source.name, e)

    def stop_event_sources(self):
        """stop the event sources"""
        for event_source in self._event_sources:
            try:
                logger.info('#Dispatcher stoping event source %s...', event_source)
                event_source.stop()
                logger.debug('#Dispatcher event source %s stopped', event_source)
            except Exception as e:
                logger.warning('#Dispatcher Stop event source %s failed: %s', event_source.name, e)

    def get_next_event(self):
        """return next event to process"""
        raise NotImplementedError

    def setup(self):
        """setup before dispatch,
        need to start all the event sources and can also do other setup things."""
        self.start_event_sources()

    def cleanup(self):
        """cleanup the dispatcher,
        stop all the evnent source and do other cleanup things.
        """
        self.stop_event_sources()

    def dispatch(self):
        """dispatch the event loop to run."""
        if not self._event_sources:
            self._stop_flag = True
            logger.warning('#Dispatcher No EventSource registered, system will exit directly.')
            return None

        # system initialized
        self.eq_put(Event(self.env, SysEvents.SYSTEM_STARTED))

        self.setup()
        last_event = None
        while True:
            event = self.get_next_event()
            if not event:
                logger.info('#Dispatcher get_next_event return None, exit.')
                break
            if event.expire != 0 and self.now.timestamp() > event.timestamp + event.expire:
                logger.warning(
                    '#Dispatcher Event %s process too slow, currend time: %s, skipped event: %s.',
                    last_event,
                    self.now,
                    event)
                self._event_process_timestamp = time.time()
                continue
            self._event_timestamp = event.timestamp
            self._event_process_timestamp = time.time()
            logger.debug('#Dispatcher Start process event: %s', event)
            for call_func in self._event_subscribes[event.name]:
                call_func(event)
                if self._stop_flag:
                    break
            last_event = event
            if self.now >= self._end_dt:
                logger.info('#Dispatcher Run out of time, dispatch finish.')
                break

        # system is exiting
        self.eq_put(Event(self.env, SysEvents.SYSTEM_EXITING))
        logger.info('#Dispatcher loop exit.')
        self._stop_flag = True
        self.cleanup()

    async def _wrap_coroutine(self, coroutine, *args, **kwargs):
        """wrap coroutine, catch exceptions"""
        # TODO: add retry and delay retry options
        exc_retry_cnt = 1
        while exc_retry_cnt > 0:
            exc_retry_cnt -= 1
            try:
                await coroutine(*args, **kwargs)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                logger.error('#Dispatcher coroutine: %s exception: %s', coroutine, e)


class BackTrackDispatcher(BaseDispatcher):
    """BackTrackDispatcher

    Simulately dispatch history events.
    it generate events when calling `get_next_event` and the event queue is empty.

    All the event generate, event dispatch,
    event callbacks are run in one process and thread serialy.

    :param start_dt: dispatch start time
    :param end_dt: dispatch end time
    :param time_step: time step for backtrack dispatch
    :param trace_process_time: pass
    """
    def __init__(self, *args, **kwargs):
        eventq = PriorityQueue()
        super().__init__(eventq=eventq, *args, **kwargs)

        self._start_dt = kwargs.get('start_dt')
        self._time_step = kwargs.get('time_step', 86400)
        self._trace_process_time = kwargs.get('trace_process_time', False)

        self._event_timestamp = self._start_dt.timestamp()
        self._event_process_timestamp = time.time()
        self._gen_dt = self.now

    @property
    def now(self):
        """datetime of current logic time"""
        if not self._trace_process_time:
            return datetime.fromtimestamp(self._event_timestamp)

        elasped_time = time.time() - self._event_process_timestamp
        return datetime.fromtimestamp(self._event_timestamp + elasped_time)

    def gen_events(self):
        """generate events from event sources."""
        futures = []
        limit_dt = datetime.fromtimestamp(self._gen_dt.timestamp() + self._time_step)
        if limit_dt > self._end_dt:
            limit_dt = self._end_dt
        for event_source in self._event_sources:
            # futures.append(event_source.gen_events(self._eventq, limit_dt))
            futures.append(self._wrap_coroutine(event_source.gen_events, self._eventq, limit_dt))

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError as e:
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()
        # loop.run_until_complete(asyncio.gather(*futures))
        loop.run_until_complete(asyncio.gather(*futures, return_exceptions=True))
        logger.debug('#Dispatcher Forward backtrack time from %s to %s...', self._gen_dt, limit_dt)
        self._gen_dt = limit_dt

    def get_next_event(self):
        """return next event to process"""
        if self.eq_size() == 0:
            self.gen_events()
            if self.eq_size() == 0:
                return None
        return self.eq_get()


class LiveDispatcher(BaseDispatcher):
    """LiveDispatcher

    Dispatch realtime events.
    all the event sources are run in one dedicated thread in `coroutine` mode.
    the event sources generate events and put them into event queue,
    the main thread get events from event queue and dispath them.


    :param stop_check_interval: stop check interval
    """
    def __init__(self, *args, **kwargs):
        eventq = Queue()
        # the thread should check exit after main thread call join, other will run in deadlock.
        self._thread_event = None
        self._thread_loop = None
        self._event_source_thread = None
        # never stop by runout time.
        end_dt = datetime.strptime('9999-12-31', '%Y-%m-%d')
        super().__init__(eventq=eventq, end_dt=end_dt, *args, **kwargs)

    def quit_handler(self, signum, frame):
        super().quit_handler(signum, frame)
        if self._thread_event:
            self._thread_loop.call_soon_threadsafe(self._thread_event.set)

    async def stop_check(self):
        """check if the dispatcher should stop, runs in the event source thread."""
        await self._thread_event.wait()
        try:    # Python <3.7 has no the method
            for task in asyncio.all_tasks():
                task.cancel()
        except AttributeError as e:
            pass
        loop = asyncio.get_event_loop()
        loop.stop()
        loop.close()

    def gen_events(self):
        """generate events from event sources."""
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop = asyncio.get_event_loop()
        self._thread_loop = loop
        self._thread_event = asyncio.Event()
        self.start_event_sources()
        while True:
            if self._stop_flag:
                break
            tasks = [self.stop_check()]
            for event_source in self._event_sources:
                # tasks.append(event_source.gen_events(self._eventq))
                tasks.append(self._wrap_coroutine(event_source.gen_events, self._eventq))
            if not tasks:
                logger.warning('#Dispatcher No event source tasks found, '
                               'EventSourceManager thread exiting.')
                break
            try:
                # loop.run_until_complete(asyncio.gather(*tasks))
                loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
            except RuntimeError as e:
                logger.error('#Dispatcher run exception: %s', e)
        logger.debug('#Dispatcher gen event thread exit.')
        return None

    def get_next_event(self):
        """get next event will in the queue to process

        :return: next event to process
        """
        if self._stop_flag or not self._event_source_thread.is_alive():
            return None
        while True:
            try:
                return self.eq_get(timeout=1)
            except Empty:
                if self._stop_flag or not self._event_source_thread.is_alive():
                    return None

    def setup(self):
        """setup before dispatch, should start all event sources"""
        self._event_source_thread = threading.Thread(
            target=self.gen_events, name='EventSourceManager')
        self._event_source_thread.start()

    def cleanup(self):
        """cleanup the dispatcher"""
        super().cleanup()
        if self._event_source_thread:
            self._event_source_thread.join()



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
runtime Environment

**this module MUSTNOT depend on any submodules of finchan, just avoid import loop.**

the `env` object is a global singleton object,
other modules/extensions can access by just import it and can access/set attributes to `env` .
"""
import logging
from datetime import datetime
from dateutil.parser import parse as parse_dt


logger = logging.getLogger(__name__)


class Env(object):
    """Global environment"""

    def __init__(self):
        self.run_mode = None
        self.options = None
        self._dispathcer = None
        self._ext_manager = None
        self._ext_objs = {}

    @property
    def now(self):
        """current datetime, wraps for dispatcher.now

        log is not permitted
        """
        if not self._dispathcer:
            return datetime.now()
        return self._dispathcer.now

    @property
    def dispatcher(self):
        """the dispatcher manager object"""
        return self._dispathcer

    @property
    def ext_manager(self):
        """the extension manager object"""
        return self._ext_manager

    def set_dispatcher(self, dispatcher):
        """set dispatcher object"""
        # from .dispatcher import Dispatcher
        # if not isinstance(dispatcher, Dispatcher):
        #     logger.warning('set env dispathcer failed, dispatcher is invalid: %s', dispatcher)
        #     return False
        self._dispathcer = dispatcher
        return True

    def set_ext_manager(self, ext_manager):
        """set extension manager object"""
        # from .exts import ExtManager
        # if not isinstance(ext_manager, ExtManager):
        #     logger.warning('set env ext_manager failed, dispatcher is invalid: %s', ext_manager)
        #     return False
        self._ext_manager = ext_manager
        return True

    def set_ext_obj(self, ext, obj):
        """set object for ext

        :param ext: name of the extension.
        :param obj: the object that call be access globaly by all.
        """
        self._ext_objs[ext] = obj

    def get_ext_obj(self, ext):
        """get object for ext that setted.

        :param ext: name of the extension.
        :return: the obj that the extension setted.
        """
        return self._ext_objs.get(ext, None)

    def get_ext_options(self, ext):
        """get options for extension named `ext` in configure file"""
        if not self.options:
            return {}
        if self.run_mode == "backtrack":
            ext_dict = self.options.get("backtrack_exts", None)
        else:
            ext_dict = self.options.get("live_track_exts", None)
        if not ext_dict:
            return {}
        return ext_dict.get(ext, {})

    def run(self):
        """wraps dispatcher's run"""
        if not self._dispathcer:
            logger.warning("No dispatcher setted, cannot run.")
            return None
        return self._dispathcer.dispatch()

    def load_exts(self, *exts):
        """load all the extensions in \*exts list"""
        if not self._ext_manager:
            logger.warning("call load_exts without setting ext_manager, no exts loaded")
        self._ext_manager.load_exts(*exts)


# env = Env()



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time


def get_id_gen(start=int(time.time()), prefix=''):
    """return a generator for str id,

    just increase 1 every time with a optional prefix."""
    i = start
    while True:
        yield '{}-{}'.format(prefix, i)
        i += 1



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extension manager"""
import os
import sys
import logging
from importlib import import_module
from collections import OrderedDict


logger = logging.getLogger(__name__)


class add_to_syspath(object):
    """A context for add a directory to sys.path for a second."""

    def __init__(self, path, prepend=True):
        self.added = False
        self.path = path
        self.prepend = prepend

    def __enter__(self):
        if self.path and self.path not in sys.path:
            if self.prepend:
                sys.path.insert(0, self.path)
            else:
                sys.path.append(self.path)
            self.added = True

    def __exit__(self, type, value, traceback):
        if self.added:
            try:
                sys.path.remove(self.path)
            except ValueError:
                pass
        # Returning False causes any exceptions to be re-raised.
        return False


class ExtManager(object):
    """Extension manager for finchan.

    An finchan extension is an importable Python module that has
    a function with the signature::

        def load_finchan_ext(env):
            # Do setup

    This function is called after your extension is imported.
    \*args and \*\*kwargs is passed from configure file's `config.live_track_exts` or
    `config.backtrack_exts` 's extension module name section depend on the run mode.

    You can also optionally define an :func:`unload_finchan_ext()`
    function, which will be called if the user unloads the extension.

    You can put your extension modules anywhere you want, as long as
    they can be imported by Python's standard import mechanism.  However,
    to make it easy to write extensions, you can also put your extensions
    in a configured path `config.ext_path`.
    This directory is added to ``sys.path`` automatically.
    """

    def __init__(self, env, ext_path="", **kwargs):
        self.env = env
        self._ext_path = ext_path
        self._loaded_exts = OrderedDict()

    def load_exts(self, exts):
        """load exts for finchan:

        :param exts: exts dict, key is extension name, value is extension kwargs
        """
        logger.info("#Extm loading exts...")
        clean_exts = []
        ext_groups = self.env.options.get("ext_groups", [])
        for ext in exts:
            if ext in ext_groups:
                clean_exts.extend(ext_groups[ext])
            else:
                clean_exts.append(ext)

        with add_to_syspath(self._ext_path):
            for ext_name in clean_exts:
                logger.info("#Extm loading ext %s...", ext_name)
                if ext_name in self._loaded_exts:
                    logger.warning(
                        "Ext[%s] has been loaded: %s, skip...",
                        ext_name,
                        self._loaded_exts[ext_name],
                    )
                    continue
                ext = import_module(ext_name)
                if not hasattr(ext, "load_finchan_ext"):
                    logger.warning("Invalid ext[%s] for finchan: %s", ext_name, ext)
                    continue
                try:
                    ext.load_finchan_ext(self.env)
                except Exception as e:
                    logger.error("Load extensiong %s failed: %s", str(e))
                    continue
                self._loaded_exts[ext_name] = ext
                logger.info("#Extm ext[%s] loaded.", ext_name)

    def unload_exts(self, *exts):
        """unload exts for finchan

        :param exts: ext_name args list
        """
        logger.info("#Extm unloading exts...")
        for ext_name in exts:
            logger.info("#Extm unloading ext %s...", ext_name)
            if ext_name not in self._loaded_exts:
                logger.warning("Ext[%s] has not been loaded, skip...", ext_name)
                continue
            ext = self._loaded_exts[ext_name]
            if not hasattr(ext, "unload_finchan_ext"):
                logger.warning("Ext[%s] has no unload func, skip...", ext_name)
                continue
            ext.unload_finchan_ext(self.env)
            del self._loaded_exts[ext_name]
            logger.info("#Extm ext[%s] unloaded.", ext_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

# 'setup.py publish' shortcut.
# if sys.argv[-1] == 'publish':
#     os.system('python setup.py sdist bdist_wheel')
#     os.system('twine upload dist/*')
#     sys.exit()

with open("README.rst", encoding="utf-8") as readme_file:
    readme = readme_file.read()

about = {}
with open(os.path.join(here, "finchan", "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), about)

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=find_packages(where=here, exclude=("tests",)),
    package_data={"": ["LICENSE"]},
    package_dir={"finchan": "finchan"},
    include_package_data=True,
    python_requires=">=3.6*",
    entry_points={"console_scripts": ["finchan=finchan.__main__:main"]},
    install_requires=["click", "uvloop", "python-dateutil", "ruamel.yaml"],
    license=about["__license__"],
    zip_safe=True,
    keywords="finchan",
    classifiers=[
        "Framework :: Robot Framework",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Chinese (Simplified)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
    ],
)



# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timer event source, interface inspired by `schedule <https://github.com/dbader/schedule>`_

usage::

    def setup(env)
        scheduler = env.get_ext_obj('finchan.exts.timer_source')
        scheduler.run_every(3).to(5).minutes().do(timer_call)
        scheduler.run_every(1).days.offset('09:31').tag('daily_task').do(timer_call)
        # runs on every two weeks' friday 09:31
        scheduler.run_every(2).weeks.offset('friday 09:31').tag('daily_task').do(timer_call)
        # runs on every month's 09 day 09:31
        scheduler.run_every(1).months.offset('09 09:31').tag('daily_task').do(timer_call)


    def clean(env)
        scheduler = env.get_ext_obj('finchan.exts.timer_source')
        scheduler.cancel_all_jobs()

"""
import random
import asyncio
import logging
import collections
from datetime import datetime
from dateutil.parser import parse as parse_dt
from dateutil.relativedelta import relativedelta as timedelta

from finchan.event import Event
from finchan.utils import get_id_gen
from finchan.interface.event_source import AbsEventSource


# name of the extension
ext_name = 'finchan.exts.timer_source'
# required extension
required_exts = []

logger = logging.getLogger(__name__)


def event_callback(event):
    """call timer func in event callback"""
    func = event.kwargs['func']
    return func(event.env, *event.kwargs['args'], **event.kwargs['kwargs'])


class JobMananger(object):
    """
    Objects instantiated by the :class:`JobMananger <JobMananger>` are
    factories to create jobs, keep record of scheduled jobs and
    handle their execution.

    TODO: schedule with asyncio.sleep(idle_seconds) and loop.create_task and task.cancel.
    """
    def __init__(self, env):
        self.env = env
        self.jobs = []
        self.curr_job = None

    def run_once(self, dt):
        """Schedule a new job that only runs once time.

        :param dt: the `datetime <datetime.datetime>` object that the job will run.
        :return: An unconfigured :class:`Job <Job>`
        """
        job = Job(step='once', next_run=dt, job_manager=self, unit='once')
        return job

    def run_every(self, step=1):
        """
        Schedule a new periodic job.

        :param step: A quantity of a certain time unit
        :return: An unconfigured :class:`Job <Job>`
        """
        job = Job(step=step, job_manager=self)
        return job

    def add_job(self, job):
        """Add a new job to the `JobMananger`"""
        self.jobs.append(job)

    def cancel(self, job):
        """cancel a job's schedule"""
        try:
            self.jobs.remove(job)
        except ValueError:
            pass

    def clear(self, tag=None):
        """
        Deletes scheduled jobs marked with the given tag, or all jobs
        if tag is `None`.

        :param tag: An identifier used to identify a subset of
                    jobs to delete
        """
        if tag is None:
            del self.jobs[:]
        else:
            self.jobs[:] = (job for job in self.jobs if tag not in job.tags)

    def cancel_all_jobs(self):
        """cancel all jobs"""
        self.jobs.clear()

    def get_next_job(self):
        """Get next job to run"""
        if not self.jobs:
            return None
        return min(self.jobs)

    def idle_seconds(self):
        """idle seconds before next job should run"""
        next_job = self.get_next_job()
        return (next_job.next_run - self.env.now).total_seconds()

    async def schedule(self):
        """TODO: implement this."""
        pass


class Job(object):
    """
    A periodic job as used by :class:`JobMananger`.

    :param step: A quantity of a certain time unit
    :param job_manager: The :class:`JobMananger <JobMananger>` instance that
                      this job will register itself with once it has
                      been fully configured in :meth:`Job.do()`.
    :param unit: perioid unit, specified below.
    :param next_run: the `datetime <datetime.datetime>` that
                     the job will run, valid for `once` unit.
    :param job_id: the ID of the job object, can be None, system will generate one for you.

    Every job runs at a given fixed time interval that is defined by:

    * a :meth:`time unit second <Job.seconds>` or :meth:`time unit minite <Job.minutes>` etc.
    * a quantity of `time units` defined by `step`

    A job is usually created and returned by :meth:`JobMananger.run_every` method,
    which also defines its `interval` step, or by :meth:`JobMananger.run_once` method,
    which specified its run time.

    unit types:

        * once
        * seconds
        * minutes
        * hours
        * days
        * weeks
        * months
        * years
    """
    id_gen = get_id_gen(prefix='Job')
    unit_types = ['once', 'seconds', 'minutes', 'hours', 'days', 'weeks', 'months', 'years']

    def __init__(self, step, job_manager, unit=None, next_run=None, job_id=None):
        self.job_id = job_id
        self.unit = unit
        self.next_run = next_run
        self.job_manager = job_manager
        self.env = job_manager.env
        self.min_step = self.max_step = step

        self.last_run = None
        self.job_func = None
        self.event_kwargs = None
        self.tags = set()

        self.offset_dt = parse_dt('1970-01-01 00:00:00')

        if not self.job_id:
            self.job_id = next(Job.id_gen)

        self.event_name = 'TimerSource.%s' % (self.job_id)
        self.event_kwargs = {}

    def to(self, step):
        """
        Schedule the job to run at an irregular (randomized) step.

        The job's interval will randomly vary from the value given
        to  `run_every <JobMananger.run_every>` to `step`.
        The range defined is inclusive on both ends.
        For example, `every(A).to(B).seconds` executes
        the job function every N seconds such that A <= N <= B.

        :param step: Maximum interval between randomized job runs
        :return: The invoked job instance
        """
        self.max_step = int(step)
        assert self.max_step >= self.min_step
        return self

    def seconds(self):
        """run in second steps

        :return: The job instance
        """
        self.unit = 'seconds'
        return self

    def minutes(self):
        """run in minute steps

        :return: The job instance
        """
        self.unit = 'minutes'
        return self

    def hours(self):
        """run in hour steps

        :return: The job instance
        """
        self.unit = 'hours'
        return self

    def days(self):
        """run in day steps

        :return: The job instance
        """
        self.unit = 'days'
        return self

    def weeks(self):
        """run in week steps

        :return: The job instance
        """
        self.unit = 'weeks'
        return self

    def months(self):
        """run in month steps

        :return: The job instance
        """
        self.unit = 'months'
        return self

    def years(self):
        """run in year steps

        :return: The job instance
        """
        self.unit = 'years'
        return self

    def tag(self, *tags):
        """
        Tags the job with one or more unique indentifiers.

        Tags must be hashable. Duplicate tags are discarded.

        :param tags: A unique list of ``Hashable`` tags.
        :return: The invoked job instance
        """
        if not all(isinstance(tag, collections.Hashable) for tag in tags):
            raise TypeError('Tags must be hashable')
        self.tags.update(tags)
        return self

    def offset(self, offset_dt=None):
        """
        Schedule the job every step hours/days/weeks/months at a specific offset.

        :param offset_dt: a `datatime` str, like '1970-01-01 00:00:00', 'monday',
                          '09:32', '07-21', '21 14:12' and so on,
                          which must can be parsed by `dateutil.parser.parse`

        picks:

            * for hours: pick the offset_dt's minite and second field.
            * for days: pick the offset_dt's time field.
            * for weeks: pick the offset_dt's time field and weekday field.
            * for months: pick the offset_dt's time field and day field.

        :return: The invoked job instance
        """
        if self.unit == 'once':
            logger.warning('#Scheduler No offset should be set for timer runs only once.')
            return

        if not offset_dt:
            logger.warning('#Scheduler Call offset but no offset_dt specified.')
            return

        try:
            self.offset_dt = parse_dt(offset_dt)
        except (ValueError, TypeError):
            logger.warning('#Scheduler offset for time in invalid: %s, set to default: %s',
                           offset_dt,
                           self.offset_dt)

    def do(self, func, *args, **kwargs):
        """
        Specifies the func that should be called every time the job runs.

        Any additional arguments are passed on to func when the job runs.

        Actually, the the func is wrapped in an event callback function,
        when it's time to call the function, the `TimerSource` just generate the event,
        the function will be executed the the `Dispatcher.dispatch` dispatch the event.

        :param func: The function to be scheduled
        :return: The invoked job instance
        """
        assert self.unit in Job.unit_types
        self.event_kwargs.update({
            'func': func,
            'args': args,
            'kwargs': kwargs
            })

        self._schedule_first_run()
        logger.debug('#Scheduler Job do last_run: %s next_run: %s', self.last_run, self.next_run)
        self.job_manager.add_job(self)
        logger.info('#Scheduler New timer Job added: %s', self)
        self.job_manager.env.dispatcher.subscribe(self.event_name, event_callback)
        return self

    def gen_event(self):
        """Generate the event and schedule the next time of the timer event occurs."""
        event = Event(self.env, self.event_name, dt=self.next_run, **self.event_kwargs)
        if self.unit == 'once':
            self.job_manager.cancel(self)
        else:
            self.last_run = self.next_run
            self._schedule_next_run()

        logger.debug('#Scheduler generate new timer event %s, job last run: %s next run: %s',
                     event,
                     self.last_run,
                     self.next_run)

        return event

    def __lt__(self, other):
        """
        PeriodicJobs are sortable based on the scheduled time they
        run next.
        """
        return self.next_run < other.next_run

    def _schedule_first_run(self):
        """Compute the instant when this job should run the first time."""
        if self.max_step is not None:
            step = random.randint(self.min_step, self.max_step)
        else:
            step = self.min_step

        replay_dict = {}
        offset_dict = {}
        env = self.job_manager.env
        if self.offset_dt.year < env.now.year:
            replay_dict['year'] = env.now.year

        if self.unit == 'once':
            return
        if self.unit == 'weeks':
            offset_dict['days'] = step * 7
        else:
            offset_dict[self.unit] = step

        if self.unit == 'minutes':
            replay_dict['second'] = self.offset_dt.second
        if self.unit == 'hours':
            replay_dict['minute'] = self.offset_dt.minute
            replay_dict['second'] = self.offset_dt.second
        if self.unit == 'days':
            replay_dict['hour'] = self.offset_dt.hour
            replay_dict['minute'] = self.offset_dt.minute
            replay_dict['second'] = self.offset_dt.second
        if self.unit == 'months':
            replay_dict['day'] = self.offset_dt.day
            replay_dict['hour'] = self.offset_dt.hour
            replay_dict['minute'] = self.offset_dt.minute
            replay_dict['second'] = self.offset_dt.second
        if self.unit == 'years':
            replay_dict['month'] = self.offset_dt.month
            replay_dict['day'] = self.offset_dt.day
            replay_dict['hour'] = self.offset_dt.hour
            replay_dict['minute'] = self.offset_dt.minute
            replay_dict['second'] = self.offset_dt.second

        bench_dt = env.now.replace(**replay_dict)
        self.next_run = bench_dt + timedelta(**offset_dict)
        while self.next_run < env.now:
            self.last_run = self.next_run
            self._schedule_next_run()

    def _schedule_next_run(self):
        """Compute the instant when this job should run next."""
        if self.unit == 'once':
            return True

        if self.max_step is not None:
            step = random.randint(self.min_step, self.max_step)
        else:
            step = self.min_step
        self.next_run = self.last_run + timedelta(**{self.unit: step})


class LiveTimerSource(AbsEventSource):
    """Live TimerEventSource, used for ``live`` mode"""
    _name = "LiveTimerEventSource"

    def __init__(self, env, *args, **kwargs):
        """init the event source"""
        self.env = env
        super().__init__(*args, **kwargs)
        self.job_manager = JobMananger(env)

    @property
    def name(self):
        """Name of the event_source"""
        return self._name

    async def gen_events(self, event_queue, limit_dt=None):
        """Generate new events for `Dispatcher`

        put all the generated/received or newly received events to
        event queue by call `Dispatcher.eq_put` , should not be blocked.
        only call by `Dispatcher` in **inner** mode, in **multithread** mode,
        the event source should run in parallel mode, and put the event to event queue.

        event_queue: put all the generated event to the event_queue
        limit_dt: for backtest, gen events before the limit_dt
        """
        while True:
            next_job = self.job_manager.get_next_job()
            if next_job and next_job.next_run <= self.env.now:
                event_queue.put(next_job.gen_event())
                if next_job.unit == 'once':
                    self.job_manager.cancel(next_job)
            else:
                # logger.debug('#Scheduler no job yet, sleep 1s.')
                await asyncio.sleep(1)

    def start(self):
        """initialize the event_source, start generate/receive events

        event_q: the event queue to put events to.
        """
        pass

    def stop(self):
        """stop the event_source, stop generate/receive events"""
        logger.info('#Scheduler timer source stopping, will clear all jobs.')
        self.job_manager.cancel_all_jobs()
        logger.debug('#Scheduler timer source stoped.')


class BackTimerSource(AbsEventSource):
    """Backtrack TimerEventSource, used for ``backtrack`` mode"""
    _name = "BackTrackTimerEventSource"

    def __init__(self, env, *args, **kwargs):
        """init the event source"""
        self.env = env
        self.job_manager = JobMananger(env)

    @property
    def name(self):
        """Name of the event_source"""
        return self._name

    async def gen_events(self, event_queue, limit_dt=None):
        """Generate new events for `Dispatcher`

        put all the generated/received or newly received events to
        event queue by call `Dispatcher.eq_put` , should not be blocked.
        only call by `Dispatcher` in **inner** mode, in **multithread** mode,
        the event source should run in parallel mode, and put the event to event queue.

        event_queue: put all the generated event to the event_queue
        limit_dt: for backtest, gen events befor the limit_dt
        """
        while True:
            next_job = self.job_manager.get_next_job()
            if next_job and next_job.next_run > limit_dt:
                break
            event_queue.put(next_job.gen_event())
            if next_job.unit == 'once':
                self.job_manager.cancel(next_job)

    def start(self):
        """initialize the event_source, start generate/receive events

        event_q: the event queue to put events to.
        """
        pass

    def stop(self):
        """stop the event_source, stop generate/receive events"""
        pass


def load_finchan_ext(env, *args, **kwargs):
    if env.run_mode == 'backtrack':
        timer_source = BackTimerSource(env)
    else:
        timer_source = LiveTimerSource(env)
    env.set_ext_obj(ext_name, timer_source.job_manager)
    env.dispatcher.register_event_source(timer_source)


def unload_finchan_ext(env):
    timer = env.timer
    env.timer = None
    timer.cancel_all_jobs()
    env.dispatcher.deregister_event_source(timer)


if __name__ == '__main__':
    # tests
    pass
